function Segs = ml_findMulti1DSegs_seq2(pixVals, nSeg, szRng, thres)
% function Segs = ml_findMulti1DSegs_seq2(pixVals, nSeg, szRng, thres)
% Finding multiple 1D segments from a signal such that the sum of
% elements in those segments is minimized.
% This function finds the segments sequentially and does not guarantee to find
% the global optimum.
% Inputs:
%   pixVals: a imH*1 vector
%   nSeg: number of segments to extract. The default is 1. If nSeg is -1,
%       segments are sequentially extracted until any newly addition will
%       carry a weighted histogram > thres. thres is another input param of
%       this function.
%	szRng: a 2*1 vector for [minLength; maxLength], range for length of
%       the individual segment
%   thres: see explanation for nSeg.
% Output:
%   Segs: a 3*K matrix with K <= nSeg. Segs([1,2],i) marks the left and
%   right ends of the interval. Segs(3,i) is the value of the interval.
% Examples:
%   pixVals = ones(40,1);
%   pixVals([4:7, 10:15, 25:35]) = -1;
%   ml_findMulti1DSegs_seq2(pixVals, 3)
%   ml_findMulti1DSegs_seq2(pixVals, 2)
%   ml_findMulti1DSegs_seq2(pixVals, 1)
% See also:
%   ml_findMulti1DSegs_seq.m: serves similar purpose but uses dif input format 
%   ml_findMulti1DSegs.m, ml_findMulti1DSegs2.m: which do guarantee to find 
%       global minimum, but cannot incorporate size constraints. 
%   
% By: Minh Hoai Nguyen (minhhoai@gmail.com)
% Date: 6 Aug 2009

imW = length(pixVals);

if size(pixVals, 2) ~= 1
    error('pixVals must be a column vector');
end;


if exist('szRng','var') && ~isempty(szRng)
    if (size(szRng, 1) ~= 2) || (size(szRng, 2) ~= 1)
        error('szRng should be a 2*1 vector');
    end;

    if (szRng(1) > szRng(2)) ||  szRng(2) <= 0 || szRng(1) > imW
        error('invalid size range');
    end;
else
    szRng = [0; imW];
end;

if ~exist('thres', 'var') || isempty(thres)
    thres = 0;
end;

Segs = m_mexFindMulti1DSegs_seq2(pixVals, nSeg, szRng, thres);



