/*
 * m_NormMulti1DSegs.cpp
 *
 * Created on: Aug 12, 2009
 * Author: Minh Hoai Nguyen, Carnegie Mellon University
 * Email:  minhhoai@cmu.edu, minhhoai@gmail.com
 */


#include "m_malloc.h"
#include "m_utils.h"
#include "m_Multi1DSegs.h"
#include "m_NormMulti1DSegs.h"

/**
 * Given 1D signal pixVals, test if:
 * \f$ \min\frac{\sum_{i \in S}pixVals(i) + \alpha}{length(S) + \beta} \leq tstVal \f$
 * \return true if the the minimum objective is <= tstVal
 */
bool isFeasible(double *& pixVals, int length, int nSeg,
      double alpha, double beta, double tstVal, Multi1DSegs *feasSegs){
   double *updatePixVals = new double[length];

   for (int i=0; i < length; i++) updatePixVals[i] = pixVals[i] - tstVal;

   double ***integralIms = cmpIntegralIms4(&updatePixVals, 1, length);
   double geVal = tstVal*beta - alpha;
   Multi1DSegs retSegs = findMulti1DSegs_helper2(length, integralIms, nSeg, geVal);


   feasSegs->nSeg   = retSegs.nSeg;
   feasSegs->length = retSegs.length;


   double energy = 0;

   feasSegs->segs = buildMatrix<double>(3, feasSegs->nSeg);
   for (int i=0; i < feasSegs->nSeg; i++){
      feasSegs->segs[0][i] = retSegs.segs[0][i];
      feasSegs->segs[1][i] = retSegs.segs[1][i];

      int sL = feasSegs->segs[1][i] - feasSegs->segs[0][i] + 1;
      feasSegs->segs[2][i] = retSegs.segs[2][i] + tstVal*sL;
      energy += feasSegs->segs[2][i];
   }

   feasSegs->energy = energy;

   delete [] updatePixVals;
   destroyArrayThree<double>(integralIms, 1, length, 2);
   double normEnergy = (energy + alpha)/(feasSegs->length + beta);
   return (normEnergy <= tstVal);
}


NormMulti1DSegs findNormMulti1DSegs(double *pixVals, int length, int nSeg,
      double alpha, double beta, double tol){

   if (beta <= 0){
      printf("findNormMulti1DSegs: beta must be positive\n");
      throw -1;
   }

   double totalSum = 0;
   double negSum = 0;
   for (int i=0; i < length; i ++){
      totalSum += pixVals[i];
      if (pixVals[i] < 0) negSum += pixVals[i];
   }

   double ub = min(alpha/beta, (totalSum + alpha)/(length + beta)); // upper bound
   double lb = (negSum + alpha)/beta;


   int iter = 0;
   while (ub - lb > tol){
      iter++;
      //printf("iter %d, lb: %g, ub: %g\n", iter, lb, ub);
      double mid = (ub + lb)/2;
      Multi1DSegs retSegs;
      if (isFeasible(pixVals, length, nSeg, alpha, beta, mid, &retSegs)){
         ub = min(mid, (retSegs.energy + alpha)/(retSegs.length + beta));
      } else lb = mid;
   }
   NormMulti1DSegs bestSegs;
   isFeasible(pixVals, length, nSeg, alpha, beta, ub, &bestSegs);
   bestSegs.nIter = iter;
   return bestSegs;
}
