#include <iostream>
#include <fstream>
#include <cmath>
#include <queue>
#include <string>
#include <sstream>
#include "m_malloc.h"
#include "m_utils.h"
#include "time.h"
#include "m_TreeNode1e.h"
#include "m_TreeNode1f.h"
#include "m_BranchBound.cpp"

/**
 * Create a tree of integral images, 2 integral images at each level
 * Inputs:
 *    imH, imW: height and width of the image.
 *    weightss_: a 2D array for several [pow(nBranch_, nLevel_ + 1)- 1]/(nBranch_ - 1)
 *       weight vectors. The first vector is the for root level.
 *       The tree has nBranch_, the next [pow(nBranch_, nLevel_)- 1]/(nBranch_ - 1) 
 *       weight vectors are for the first branch and so on so forth. 
 *    nBranch_: number of branches
 *    nLevel_: number of levels
 */
IntImTree::IntImTree(int **kpAssign, int nKeyPnt, int imH_, int imW_,
                     double **weightss_, int nHistBin_, 
                     int nBranch_, int nLevel_){
   imH = imH_;
   imW = imW_;
   nLevel = nLevel_;
   nBranch = nBranch_;
   sqrtBranch = (int)sqrt((double)nBranch);
   
   integralIms = cmpIntegralIms3(kpAssign, nKeyPnt, imH, imW, weightss_[0], nHistBin_); 

   if (nLevel == 0){
      isLeaf = true;        
   } else {
      isLeaf = false;
      int k = (int)(pow((double)nBranch, nLevel) - 1)/(nBranch - 1);
      for (int i=0; i < nBranch; i++){
         branches.push_back(new IntImTree(kpAssign, nKeyPnt, imH, imW, &weightss_[1+i*k], nHistBin_,
            nBranch, nLevel - 1));         
      }     
   }
};


double
IntImTree::cmpLb(WindowPair winPair){
   Window smallestWin = winPair.getSmallestWin();  
   double *hist_min = smallestWin.getPosNegScore(integralIms);
   Window largestWin = winPair.getLargestWin();
   double *hist_max = largestWin.getPosNegScore(integralIms);
   double lb = hist_min[0] - hist_max[1];

   destroyVector<double>(hist_min, 2);
   destroyVector<double>(hist_max, 2);
   int x, y;
   if (!isLeaf){
      for (int i=0; i < nBranch; i++){         
         y = i % sqrtBranch + 1;
         x = i/sqrtBranch + 1;
         WindowPair brWinPair = winPair.getSubWinPair(x, y, sqrtBranch, sqrtBranch);         
         lb += branches[i]->cmpLb(brWinPair);
      }
   } 
   return lb;
} 


double
IntImTree::cmpE(Window win){
   if (win.lr_x < 0){
      return 0;
   } 
   double energy = 0;
   double *hist_ = win.getPosNegScore(integralIms);
   energy = hist_[0] - hist_[1]; 
   destroyVector<double>(hist_, 2);
   
   int x, y;
   if (!isLeaf){
      for (int i=0; i < nBranch; i++){         
         y = i % sqrtBranch + 1;
         x = i/sqrtBranch + 1;
         Window brWin = win.getSubWin(x, y, sqrtBranch, sqrtBranch);
         energy += branches[i]->cmpE(brWin); 
      }
   }
   return energy;
}


IntImTree::~IntImTree(){
   destroyArrayThree<double>(integralIms, imH, imW, 2);
   if (!isLeaf){
      for (int i=0; i < nBranch; i++){                  
         delete branches[i]; 
      }
   }
};

void TreeNode1f::cmpBnds(){
   // Test if there is a window in the range.
   Res xx, yy;
   if ((!findEnds(Res(winPair.win1.ul_x, winPair.win1.lr_x),
        Res(winPair.win2.ul_x, winPair.win2.lr_x),
        Res(szRange.wmin, szRange.wmax), &xx)) ||
      (!findEnds(Res(winPair.win1.ul_y, winPair.win1.lr_y),
        Res(winPair.win2.ul_y, winPair.win2.lr_y),
        Res(szRange.hmin, szRange.hmax), &yy))){
      
       lb = INF;
       ub = INF;
       return;
   }

   // window in the range for the upper bound energy.
   ubWin = Window(xx.x, yy.x, xx.y, yy.y);
   ub = intImTree->cmpE(ubWin); 
   lb = intImTree->cmpLb(winPair); 
}

IntImTree *TreeNode1f::intImTree = NULL;
SizeRange TreeNode1f::szRange;



Window findBox1f(int **kpAssign, int nKeyPnt, int imH, int imW, double **weightss, 
                 int nHistBin, int nBranch, int nLevel, SizeRange szRange, double tolFac, double *energy){
 
   IntImTree intImTree(kpAssign, nKeyPnt, imH, imW, weightss, nHistBin, nBranch, nLevel);      
   TreeNode1f::intImTree = &intImTree;
   TreeNode1f::szRange = szRange;

   Window biggestWin(0,0, imW-1, imH-1);
   TreeNode1f rootNode(WindowPair(biggestWin, biggestWin));
   TreeNode1f bestNode = m_branch_bound<TreeNode1f>(rootNode, tolFac);
   
   energy[0] = intImTree.cmpE(bestNode.ubWin); 
   return bestNode.ubWin; 
}
