/*
 * m_multiLabelKSeg.h
 *
 * Created on: Nov 29, 2009
 * Author: Minh Hoai Nguyen, Carnegie Mellon University
 * Email:  minhhoai@cmu.edu, minhhoai@gmail.com
 */

#ifndef M_MULTILABELKSEG_H_
#define M_MULTILABELKSEG_H_

#include "m_malloc.h"
#include <vector>
#include "m_utils.h"
#include <iostream>
#include<gsl/gsl_cblas.h>

using namespace std;


class MLKS_Rslt{
public:
   int startIdx, endIdx, label;
   MLKS_Rslt(int startIdx_, int endIdx_, int label_){
      startIdx = startIdx_;
      endIdx = endIdx_;
      label = label_;
   }

   MLKS_Rslt(const MLKS_Rslt& rhs){
      startIdx = rhs.startIdx;
      endIdx = rhs.endIdx;
      label = rhs.label;
   }
};

/**
 * Compute the energy
 */
double cmpMLKS_energy(vector<MLKS_Rslt> labelAssgn, double **A);


/**
 * Compute MMSeg energy
 */
double cmpMMCS_energy(double **A, int nLabel, int len, vector<MLKS_Rslt>);


/**
 * Entry class for multi-label-kSeg
 */
class MLKS_Entry{
public:
   MLKS_Entry(){
      prevDifIdx = -1;
      prevDifLabel = -1;
   };

   /**
    * This operator does nothing, it is created because buildArrayThree needs it
    */
   MLKS_Entry &operator=(const int num){
      return *this;
   };
   MLKS_Entry(const MLKS_Entry& rhs){
      prevDifIdx = rhs.prevDifIdx;
      prevDifLabel = rhs.prevDifLabel;
      f = rhs.f;
      f_bestAlt = rhs.f_bestAlt;
      bestAltLabel = rhs.bestAltLabel;
   };

   double f, f_bestAlt;
   int bestAltLabel;
   int prevDifIdx;
   int prevDifLabel;
};

/**
 * Entry class for multi-label-kSeg
 */
class MLKS_Entry2{
public:
   MLKS_Entry2(){prevLabel = -1;};

   /**
    * This operator does nothing, it is created because buildArrayThree needs it
    */
   MLKS_Entry2 &operator=(const int num){
      return *this;
   };
   MLKS_Entry2(const MLKS_Entry2& rhs){
      prevLabel = rhs.prevLabel;
      f = rhs.f;
   };

   double f;
   int prevLabel;
};


/**
 * Find the optimal labeling that minimizes the sum of weights
 * \f$ \min_{y} \sum_{i=0}^{len-1} A_{iy_i} \textrm{s.t.} \sum_{i=0}^{len-2} \delta(y_i \neq y_{i+1}) \leq nSeg \f$
 * \param A a nLabel*len matrix
 * \param nLabel # of different labels
 * \param len the length of the signal
 * \param nSeg maximum # of segments that [1,len] can be broken into.
 * \return a vector of MLKS_Rslt, and the returned energy is return.
 */
vector<MLKS_Rslt> multiLabelKseg(double **A, int nLabel, int len, int nSeg, double *energy);

/**
 * Find the optimal labeling that minimizes the sum of weights
 * \f$ \min_{y} \sum_{i=0}^{len-1} A_{iy_i} \textrm{s.t.} \sum_{i=0}^{len-2} \delta(y_i \neq y_{i+1}) \leq nSeg \f$
 * \param A a nLabel*len matrix
 * \param nLabel # of different labels
 * \param len the length of the signal
 * \param nSeg maximum # of segments that [1,len] can be broken into.
 * \param segSwchInd a nLabel*nLabel indicator for segment switching
 *    segSwchInd(i,j) is 1: jumping from label i to j will be considered two different segments
 *    segSwchInd(i,j) is 0: jumping from label i to j will not be considered two different segments
 *    E.g., in AU detection: 0-neutral, 1-onset, 2-offset. From label 1 to 2 doesn't mean
 *    we are switching between segments, but from 2 to 1 is.
 *    Compared with multLabelKSeg, this function is more general. The other function
 *    has segSwchInd(i,j) = 0 iff i = j. If that is the case, you should
 *    use the other function because it is a bit more efficient.
 * \return a vector of MLKS_Rslt, and the returned energy is return.
 */
int *multiLabelKseg2(double **A, int nLabel, int len, int nSeg, bool **segSwchInd, double *energy);



/**
 * Find the optimal labeling that minimizes the sum of segment weights.
 * The weight of each segment is the sum of weight inside the segment
 * The size of each sub-segment must be within lmin and lmax, inclusisve.
 * \param A a nLabel*len matrix
 * \param nLabel # of different labels
 * \param len the length of the signal
 * \param lmin, lmax: minimum and maximum segment lengths.
 * \param energy: the return energy
 * \return a vector of MLKS_Rslt.
 */
vector<MLKS_Rslt> multiLabelKseg3(double **A, int nLabel, int len, int lmin, int lmax, double *energy);


/**
 * Find the optimal labeling that minimizes the sum of weights
 * This function performs segmentation similar to multiLabelKseg; however,
 * it uses the MMC criteria instead. Furthermore, instead of constraining
 * the number of segments, it constraints the length of the segments.
 * \param A a nLabel*len matrix
 * \param nLabel # of different labels
 * \param len the length of the signal
 * \param lmin, lmax minimum and maximum lengths (inclusive) of segments
 * \param energy: the return energy
 * \return a vector of MLKS_Rslt.
 */
vector<MLKS_Rslt> MMCSeg(double **A, int nLabel, int len,
      int lmin, int lmax, double *energy);


/**
 * Return a random segmentation of a time series.
 * Each sub-segment should have length bounded by lmin and lmax.
 */
vector<MLKS_Rslt> randomSeg(int len, int nLabel, int lmin, int lmax, double *nValidSeg);

/**
 * Find the optimal labeling that optimizes k-means energy
 * @param D a n*d data matrix
 * @param d dim of the data
 * @param n length of the sequence
 * @param Cs the centroids of k-means
 * @param nClass number of class
 * @param lmin minimum segment length
 * @param lmax maximum segment length
 * @param energy k-means energy
 * @return
 */
vector<MLKS_Rslt> KmeansSeg(double **D,
      int d, int n, double **Cs, int nClass, int lmin, int lmax, bool shldNorm, double &energy);

#endif /* M_MULTILABELKSEG_H_ */
